# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling datasets."""

import concurrent.futures
import fnmatch
import imghdr
import os
import shlex
import shutil
import tempfile
import time
import uuid
from collections import OrderedDict
from contextlib import contextmanager
from pathlib import Path
from subprocess import PIPE, SubprocessError
from typing import Dict, List, Optional
from urllib import error, parse
from urllib.parse import ParseResult, unquote, urlparse
from urllib.request import urlretrieve

import attr
import patoolib
import requests
from git import GitCommandError, GitError, Repo
from wcmatch import glob
from yagup import GitURL

from renku.core import errors
from renku.core.management.clone import clone
from renku.core.management.command_builder.command import inject
from renku.core.management.config import RENKU_HOME
from renku.core.management.dataset import get_dataset
from renku.core.management.dataset.datasets_provenance import DatasetsProvenance
from renku.core.management.interface.database_dispatcher import IDatabaseDispatcher
from renku.core.management.repository import RepositoryApiMixin
from renku.core.metadata.immutable import DynamicProxy
from renku.core.models.dataset import (
    Dataset,
    DatasetFile,
    ImageObject,
    RemoteEntity,
    get_dataset_data_dir,
    is_dataset_name_valid,
)
from renku.core.models.provenance.agent import Person
from renku.core.models.provenance.annotation import Annotation
from renku.core.models.refs import LinkReference
from renku.core.utils import communication
from renku.core.utils.git import (
    add_to_git,
    find_previous_commit,
    get_oauth_url,
    get_object_hash,
    get_renku_repo_url,
    have_same_remote,
    run_command,
)
from renku.core.utils.urls import get_slug, remove_credentials


@attr.s
class DatasetsApiMixin(object):
    """Client for handling datasets."""

    POINTERS = "pointers"
    """Directory for storing external pointer files."""

    CACHE = "cache"
    """Directory to cache transient data."""

    DATASET_IMAGES = "dataset_images"
    """Directory for dataset images."""

    SUPPORTED_SCHEMES = ("", "file", "http", "https", "git+https", "git+ssh")

    _database = None

    @property
    def renku_dataset_images_path(self):
        """Return a ``Path`` instance of Renku dataset metadata folder."""
        return self.path / self.renku_home / self.DATASET_IMAGES

    @property
    def renku_pointers_path(self):
        """Return a ``Path`` instance of Renku pointer files folder."""
        path = self.path / self.renku_home / self.POINTERS
        path.mkdir(exist_ok=True)
        return path

    @property
    def datasets(self) -> Dict[str, Dataset]:
        """A map from datasets name to datasets."""
        datasets_provenance = DatasetsProvenance()
        return {d.name: d for d in datasets_provenance.datasets}

    @staticmethod
    def get_dataset(name, strict=False, immutable=False) -> Optional[Dataset]:
        """Load dataset reference file."""
        return get_dataset(name=name, strict=strict, immutable=immutable)

    @contextmanager
    @inject.autoparams("database_dispatcher")
    def with_dataset(
        self,
        database_dispatcher: IDatabaseDispatcher,
        name: str = None,
        create: bool = False,
        commit_database: bool = False,
        creator: Person = None,
    ):
        """Yield an editable metadata object for a dataset."""
        dataset = self.get_dataset(name=name)

        if dataset is None:
            if not create:
                raise errors.DatasetNotFound(name=name)

            # NOTE: Don't update provenance when creating here because it will be updated later
            dataset = self.create_dataset(name=name, update_provenance=False)
        elif create:
            raise errors.DatasetExistsError('Dataset exists: "{}".'.format(name))

        try:
            yield dataset
        except Exception:
            # TODO use a general clean-up strategy: https://github.com/SwissDataScienceCenter/renku-python/issues/736
            raise

        if commit_database:
            datasets_provenance = DatasetsProvenance()
            datasets_provenance.add_or_update(dataset, creator=creator)
            database_dispatcher.current_database.commit()

    def create_dataset(
        self,
        name=None,
        title=None,
        description=None,
        creators=None,
        keywords=None,
        images=None,
        safe_image_paths=None,
        update_provenance=True,
        custom_metadata=None,
    ):
        """Create a dataset."""
        if not name:
            raise errors.ParameterError("Dataset name must be provided.")

        if not is_dataset_name_valid(name):
            valid_name = get_slug(name)
            raise errors.ParameterError(f'Dataset name "{name}" is not valid (Hint: "{valid_name}" is valid).')

        if self.get_dataset(name=name):
            raise errors.DatasetExistsError(f"Dataset exists: '{name}'")

        if not title:
            title = name

        if creators is None:
            creators = [Person.from_git(self.repo)]

        keywords = keywords or ()

        annotations = None

        if custom_metadata:
            annotations = [Annotation(id=Annotation.generate_id(), source="renku", body=custom_metadata)]

        dataset = Dataset(
            identifier=None,
            name=name,
            title=title,
            description=description,
            creators=creators,
            keywords=keywords,
            project_id=self.project.id,
            annotations=annotations,
        )

        if images:
            safe_image_paths = safe_image_paths or []
            safe_image_paths.append(self.path)
            self.set_dataset_images(dataset, images, safe_image_paths)

        if update_provenance:
            datasets_provenance = DatasetsProvenance()
            datasets_provenance.add_or_update(dataset)

        return dataset

    def update_dataset_custom_metadata(self, dataset: Dataset, custom_metadata: Dict):
        """Update custom metadata on a dataset."""

        existing_metadata = [a for a in dataset.annotations if a.source != "renku"]

        existing_metadata.append(Annotation(id=Annotation.generate_id(), body=custom_metadata, source="renku"))

        dataset.annotations = existing_metadata

    def set_dataset_images(self, dataset: Dataset, images, safe_image_paths=None):
        """Set the images on a dataset."""
        safe_image_paths = safe_image_paths or []

        if not images:
            images = []

        image_folder = self.renku_dataset_images_path / dataset.initial_identifier
        image_folder.mkdir(exist_ok=True, parents=True)

        previous_images = dataset.images or []

        dataset.images = []

        images_updated = False

        for img in images:
            position = img["position"]
            content_url = img["content_url"]

            if any(i.position == img["position"] for i in dataset.images):
                raise errors.DatasetImageError(f"Duplicate dataset image specified for position {position}")

            existing = next(
                (i for i in previous_images if i.position == img["position"] and i.content_url == img["content_url"]),
                None,
            )

            if existing:
                dataset.images.append(existing)
                continue
            image_type = None
            if urlparse(content_url).netloc:
                # NOTE: absolute url
                if not img.get("mirror_locally", False):
                    dataset.images.append(
                        ImageObject(
                            content_url=content_url,
                            position=position,
                            id=ImageObject.generate_id(dataset, position),
                        )
                    )
                    images_updated = True
                    continue

                # NOTE: mirror the image locally
                try:
                    path, _ = urlretrieve(content_url)
                except error.URLError as e:
                    raise errors.DatasetImageError(f"Dataset image with url {content_url} couldn't be mirrored") from e

                image_type = imghdr.what(path)
                if image_type:
                    image_type = f".{image_type}"

                content_url = path
                safe_image_paths.append(Path(path).parent)

            path = content_url
            if not os.path.isabs(path):
                path = os.path.normpath(os.path.join(self.path, path))

            if not os.path.exists(path) or not any(os.path.commonprefix([path, p]) == str(p) for p in safe_image_paths):
                # NOTE: make sure files exists and prevent path traversal
                raise errors.DatasetImageError(f"Dataset image with relative path {content_url} not found")

            if not path.startswith(str(image_folder)):
                # NOTE: only copy dataset image if it's not in .renku/datasets/<id>/images/ already
                if image_type:
                    ext = image_type
                else:
                    _, ext = os.path.splitext(content_url)

                img_path = image_folder / f"{position}{ext}"
                shutil.copy(path, img_path)
            else:
                img_path = path

            dataset.images.append(
                ImageObject(
                    content_url=str(img_path.relative_to(self.path)),
                    position=position,
                    id=ImageObject.generate_id(dataset=dataset, position=position),
                )
            )
            images_updated = True

        new_urls = [i.content_url for i in dataset.images]

        for prev in previous_images:
            # NOTE: Delete images if they were removed
            if prev.content_url in new_urls or urlparse(prev.content_url).netloc:
                continue

            path = prev.content_url
            if not os.path.isabs(path):
                path = os.path.normpath(os.path.join(self.path, path))

            os.remove(path)

        return images_updated or dataset.images != previous_images

    @inject.autoparams()
    def add_data_to_dataset(
        self,
        dataset,
        urls,
        force=False,
        overwrite=False,
        sources=(),
        destination="",
        ref=None,
        external=False,
        extract=False,
        all_at_once=False,
        destination_names=None,
        repository=None,
        clear_files_before=False,
    ):
        """Import the data into the data directory."""
        dataset_datadir = get_dataset_data_dir(self, dataset)

        destination = destination or Path(".")
        destination = self._resolve_path(dataset_datadir, destination)
        destination = self.path / dataset_datadir / destination

        if destination.exists() and not destination.is_dir():
            raise errors.ParameterError(f'Destination is not a directory: "{destination}"')

        self.check_external_storage()

        files = []
        if all_at_once:  # Importing a dataset
            files = self._add_from_urls(
                urls=urls, destination_names=destination_names, destination=destination, extract=extract
            )
        else:
            for url in urls:
                is_remote, is_git, url = _check_url(url)
                if is_git and is_remote:  # Remote git repo
                    sources = sources or ()
                    new_files = self._add_from_git(
                        url=url, sources=sources, destination=destination, ref=ref, repository=repository
                    )
                else:
                    if sources:
                        raise errors.UsageError('Cannot use "--source" with URLs or local files.')

                    if not is_remote:  # Local path, might be git
                        if is_git:
                            communication.warn(
                                "Adding data from local Git repository: "
                                + "Use remote's Git URL instead to enable "
                                + "lineage information and updates."
                            )
                        u = parse.urlparse(url)
                        new_files = self._add_from_local(
                            dataset=dataset, path=u.path, external=external, destination=destination
                        )
                    else:  # Remote URL
                        new_files = self._add_from_url(url=url, destination=destination, extract=extract)

                files.extend(new_files)

        # Remove all files that are under a .git directory
        paths_to_avoid = [f["path"] for f in files if ".git" in str(f["path"]).split(os.path.sep)]
        if paths_to_avoid:
            files = [f for f in files if f["path"] not in paths_to_avoid]
            communication.warn(
                "Ignored adding paths under a .git directory:\n  " + "\n  ".join(str(p) for p in paths_to_avoid)
            )

        files_to_commit = {str(self.path / f["path"]) for f in files}

        if not force:
            ignored_files = self.find_ignored_paths(*files_to_commit)
            if ignored_files:
                ignored_files = set(ignored_files)
                files_to_commit = files_to_commit.difference(ignored_files)
                ignored_sources = []
                for file_ in files:
                    if str(self.path / file_["path"]) in ignored_files:
                        operation = file_.get("operation")
                        if operation:
                            src, _, _ = operation
                            ignored_sources.append(src)
                        else:
                            ignored_sources.append(file_["path"])

                files = [f for f in files if str(self.path / f["path"]) in files_to_commit]
                communication.warn(
                    "Theses paths are ignored by one of your .gitignore "
                    + 'files (use "--force" flag if you really want to add '
                    + "them):\n  "
                    + "\n  ".join([str(p) for p in ignored_sources])
                )

        # all files at this point can be force-added

        if not overwrite:
            existing_files = []
            for path in files_to_commit:
                relative_path = Path(path).relative_to(self.path)
                if dataset.find_file(relative_path):
                    existing_files.append(path)

            if existing_files:
                files_to_commit = files_to_commit.difference(existing_files)
                files = [f for f in files if str(self.path / f["path"]) in files_to_commit]
                communication.warn(
                    "These existing files were not overwritten "
                    + '(use "--overwrite" flag to overwrite them):\n  '
                    + "\n  ".join([str(p) for p in existing_files])
                )

        for data in files:
            operation = data.pop("operation", None)
            if not operation:
                continue

            src, dst, action = operation

            # Remove existing file if any
            self.remove_file(dst)
            dst.parent.mkdir(parents=True, exist_ok=True)

            if action == "copy":
                shutil.copy(src, dst)
            elif action == "move":
                shutil.move(src, dst, copy_function=shutil.copy)
            elif action == "symlink":
                self._create_external_file(src, dst)
                data["is_external"] = True
            else:
                raise errors.OperationError(f"Invalid action {action}")

        # Track non-symlinks in LFS
        if self.check_external_storage():
            self.track_paths_in_storage(*files_to_commit)

        # Force-add to include possible ignored files
        add_to_git(self.repo.git, *files_to_commit, force=True)
        self.repo.git.add(self.renku_pointers_path, force=True)

        staged_files = self.repo.index.diff("HEAD")
        if staged_files:
            msg = "renku dataset: committing {} newly added files".format(len(files_to_commit))
            skip_hooks = not self.external_storage_requested
            self.repo.index.commit(msg, skip_hooks=skip_hooks)
        else:
            communication.warn("No new file was added to project")

        if not files:
            return

        # Generate the DatasetFiles
        dataset_files = []
        for data in files:
            dataset_file = DatasetFile.from_path(
                client=self, path=data["path"], source=data["source"], based_on=data.get("based_on")
            )
            dataset_files.append(dataset_file)

        if clear_files_before:
            dataset.clear_files()
        dataset.add_or_update_files(dataset_files)
        datasets_provenance = DatasetsProvenance()
        datasets_provenance.add_or_update(dataset, creator=Person.from_client(self))

    def is_protected_path(self, path):
        """Checks if a path is a protected path."""
        try:
            path_in_repo = str(path.relative_to(self.path))
        except ValueError:
            return False

        for protected_path in self.RENKU_PROTECTED_PATHS:
            if fnmatch.fnmatch(path_in_repo, protected_path):
                return True

        return False

    def _add_from_local(self, dataset, path, external, destination):
        """Add a file or directory from a local filesystem."""
        src = Path(os.path.abspath(path))

        if not src.exists():
            raise errors.ParameterError(f"Cannot find file/directory: {path}")

        dst = destination / src.name

        # if we have a directory, recurse
        if src.is_dir():
            if dst.exists() and not dst.is_dir():
                raise errors.ParameterError(f'Cannot copy directory to a file: "{dst}"')
            if src == (self.path / get_dataset_data_dir(self, dataset)).resolve():
                raise errors.ParameterError(f"Cannot add dataset's data directory recursively: {path}")

            if self.is_protected_path(src):
                raise errors.ProtectedFiles([src])

            files = []
            for f in src.iterdir():
                files.extend(
                    self._add_from_local(dataset=dataset, path=os.path.abspath(f), external=external, destination=dst)
                )
            return files
        else:
            # Check if file is in the project and return it
            path_in_repo = None
            if self.is_external_file(src):
                path_in_repo = path
            else:
                try:
                    path_in_repo = src.relative_to(self.path)
                except ValueError:
                    pass
                else:
                    if self.is_protected_path(src):
                        raise errors.ProtectedFiles([src])

            if path_in_repo:
                return [{"path": path_in_repo, "source": path_in_repo, "parent": self}]

        action = "symlink" if external else "copy"
        return [
            {
                "path": dst.relative_to(self.path),
                "source": os.path.relpath(str(src), str(self.path)),
                "parent": self,
                "operation": (src, dst, action),
            }
        ]

    def _add_from_urls(self, urls, destination, destination_names, extract):
        listeners = communication.get_listeners()

        def subscribe_communication_listeners(function, **kwargs):
            try:
                for communicator in listeners:
                    communication.subscribe(communicator)
                return function(**kwargs)
            finally:
                for communicator in listeners:
                    communication.unsubscribe(communicator)

        files = []
        max_workers = min(os.cpu_count() - 1, 4) or 1
        with concurrent.futures.ThreadPoolExecutor(max_workers) as executor:
            futures = {
                executor.submit(
                    subscribe_communication_listeners,
                    self._add_from_url,
                    url=url,
                    destination=destination,
                    extract=extract,
                    filename=name,
                )
                for url, name in zip(urls, destination_names)
            }

            for future in concurrent.futures.as_completed(futures):
                files.extend(future.result())

        return files

    def _add_from_url(self, url, destination, extract, filename=None):
        """Process adding from url and return the location on disk."""
        url = self._provider_check(url)

        try:
            start = time.time() * 1e3
            tmp_root, paths = self._download(url=url, filename=filename, extract=extract)

            exec_time = (time.time() * 1e3 - start) // 1e3
            # If execution time was less or equal to zero seconds,
            # block the thread a bit to avoid being rate limited.
            if exec_time == 0:
                time.sleep(min(os.cpu_count() - 1, 4) or 1)

        except (requests.exceptions.HTTPError, error.HTTPError) as e:  # pragma nocover
            raise errors.OperationError("Cannot download from {}".format(url)) from e

        paths = [(src, destination / src.relative_to(tmp_root)) for src in paths if not src.is_dir()]
        return [
            {
                "operation": (src, dst, "move"),
                "path": dst.relative_to(self.path),
                "source": remove_credentials(url),
                "parent": self,
            }
            for src, dst in paths
        ]

    def _add_from_git(self, url, sources, destination, ref, repository=None):
        """Process adding resources from another git repository."""
        from renku import LocalClient

        u = parse.urlparse(url)

        sources = self._resolve_paths(u.path, sources)

        if not repository:
            repository, repo_path = self.prepare_git_repo(url, ref, depth=None)
        else:
            repo_path = Path(repository.working_dir)

        # Get all files from repo that match sources
        files = set()
        used_sources = set()
        for file in repository.head.commit.tree.traverse():
            path = file.path
            result = self._get_src_and_dst(path, repo_path, sources, destination, used_sources)

            if result:
                files.add(result)

        unused_sources = set(sources.keys()) - used_sources
        if unused_sources:
            unused_sources = {str(s) for s in unused_sources}
            raise errors.ParameterError("No such file or directory", param_hint=unused_sources)

        # Create metadata and move files to dataset
        results = []
        remote_client = LocalClient(repo_path)

        # Pull files from LFS
        paths = set()
        for path, src, _ in files:
            if src.is_dir():
                continue
            if src.is_symlink():
                try:
                    path = str(src.resolve().relative_to(repo_path))
                except ValueError:  # External file
                    pass
            paths.add(path)
        self._fetch_lfs_files(repo_path, paths)

        new_files = []

        for path, src, dst in files:
            if not src.is_dir():
                path_in_dst_repo = dst.relative_to(self.path)
                if path_in_dst_repo in new_files:  # A path with the same destination is already copied
                    continue

                new_files.append(path_in_dst_repo)

                if remote_client.is_external_file(src):
                    operation = (src.resolve(), dst, "symlink")
                else:
                    operation = (src, dst, "move")

                checksum = get_object_hash(repo=remote_client.repo, revision="HEAD", path=path)
                based_on = RemoteEntity(checksum=checksum, path=path, url=url)

                results.append(
                    {
                        "path": path_in_dst_repo,
                        "source": remove_credentials(url),
                        "parent": self,
                        "based_on": based_on,
                        "operation": operation,
                    }
                )

        return results

    @staticmethod
    def _ensure_dropbox(url):
        """Ensure dropbox url is set for file download."""
        if not isinstance(url, ParseResult):
            url = parse.urlparse(url)

        query = url.query or ""
        if "dl=0" in url.query:
            query = query.replace("dl=0", "dl=1")
        else:
            query += "dl=1"

        url = url._replace(query=query)
        return url

    def _provider_check(self, url):
        """Check additional provider related operations."""
        # NOTE: Check if the url is a redirect.
        r = requests.head(url, allow_redirects=True)
        url = parse.urlparse(r.url)

        if "dropbox.com" in url.netloc:
            url = self._ensure_dropbox(url)

        return parse.urlunparse(url)

    def _resolve_paths(self, root_path, paths):
        """Check if paths are within a root path and resolve them."""
        result = OrderedDict()  # Used as an ordered-set
        for path in paths:
            r = self._resolve_path(root_path, path)
            result[r] = None
        return result

    @staticmethod
    def _resolve_path(root_path, path):
        """Check if a path is within a root path and resolve it."""
        try:
            root_path = Path(root_path).resolve()
            path = os.path.abspath(root_path / path)
            return Path(path).relative_to(root_path)
        except ValueError:
            raise errors.ParameterError("File {} is not within path {}".format(path, root_path))

    @staticmethod
    def _get_src_and_dst(path, repo_path, sources, dst_root, used_sources):
        is_wildcard = False
        matched_pattern = None

        if not sources:
            source = Path(".")
        else:
            source = None
            for s in sources.keys():
                try:
                    Path(path).relative_to(s)
                except ValueError:
                    if glob.globmatch(path, str(s), flags=glob.GLOBSTAR):
                        is_wildcard = True
                        source = Path(path)
                        used_sources.add(s)
                        matched_pattern = str(s)
                        break
                else:
                    source = Path(s)
                    used_sources.add(source)
                    break

            if not source:
                return

        if is_wildcard:
            # Search to see if a parent of the path matches the pattern and return it
            while glob.globmatch(str(source.parent), matched_pattern, flags=glob.GLOBSTAR) and source != source.parent:
                source = source.parent

        src = repo_path / path
        source_name = source.name
        relative_path = Path(path).relative_to(source)

        if src.is_dir() and is_wildcard:
            sources[source] = None
            used_sources.add(source)

        dst = dst_root / source_name / relative_path

        return path, src, dst

    @staticmethod
    def _fetch_lfs_files(repo_path, paths):
        """Fetch and checkout paths that are tracked by Git LFS."""
        repo_path = str(repo_path)

        try:
            paths = [shlex.quote(p) for p in paths]
            status = run_command(
                ["git", "lfs", "pull", "--include"],
                *paths,
                separator=",",
                stderr=PIPE,
                cwd=repo_path,
                universal_newlines=True,
            )
            if status.returncode != 0:
                message = "\n\t".join(status.stderr.split("\n"))
                raise errors.GitError(f"Cannot pull LFS objects from server: {message}")
        except KeyboardInterrupt:
            raise
        except SubprocessError as e:
            raise errors.GitError(f"Cannot pull LFS objects from server: {e}")

    def move_files(self, files, to_dataset):
        """Move files and their metadata from one or more datasets to a target dataset."""
        datasets = [d.copy() for d in self.datasets.values()]
        if to_dataset:
            # NOTE: Use the same dataset object or otherwise a race happens if dataset is in both source and destination
            to_dataset: Dataset = next(d for d in datasets if d.name == to_dataset)
        modified_datasets = {}

        progress_name = "Updating dataset metadata"
        communication.start_progress(progress_name, total=len(files))
        try:
            for src, dst in files.items():
                src = src.relative_to(self.path)
                dst = dst.relative_to(self.path)
                # NOTE: Files are moved at this point, so, we use can use dst
                new_dataset_file = DatasetFile.from_path(self, dst)
                for dataset in datasets:
                    removed = dataset.unlink_file(src, missing_ok=True)
                    if removed:
                        modified_datasets[dataset.name] = dataset
                        new_dataset_file.based_on = removed.based_on
                        new_dataset_file.source = removed.source
                        if not to_dataset:
                            dataset.add_or_update_files(new_dataset_file)

                    # NOTE: Update dataset if it contains a destination that is being overwritten
                    modified = dataset.find_file(dst)
                    if modified:
                        modified_datasets[dataset.name] = dataset
                        dataset.add_or_update_files(new_dataset_file)

                if to_dataset:
                    to_dataset.add_or_update_files(new_dataset_file)

                communication.update_progress(progress_name, amount=1)
        finally:
            communication.finalize_progress(progress_name)

        datasets_provenance = DatasetsProvenance()
        for dataset in modified_datasets.values():
            datasets_provenance.add_or_update(dataset, creator=Person.from_client(self))
        if to_dataset:
            datasets_provenance.add_or_update(to_dataset, creator=Person.from_client(self))

    def update_dataset_local_files(self, records: List[DynamicProxy], delete=False):
        """Update files metadata from the git history."""
        updated_files: List[DynamicProxy] = []
        deleted_files: List[DynamicProxy] = []
        progress_text = "Checking for local updates"

        try:
            communication.start_progress(progress_text, len(records))
            for file in records:
                communication.update_progress(progress_text, 1)

                if file.based_on or file.is_external:
                    continue

                if not (self.path / file.entity.path).exists():
                    deleted_files.append(file)
                    continue

                current_checksum = get_object_hash(repo=self.repo, revision="HEAD", path=file.entity.path)
                if not current_checksum:
                    deleted_files.append(file)
                elif current_checksum != file.entity.checksum:
                    updated_files.append(file)
        finally:
            communication.finalize_progress(progress_text)

        if updated_files or (deleted_files and delete):
            self._update_datasets_metadata(updated_files, deleted_files, delete)

        return updated_files, deleted_files

    def _update_datasets_metadata(
        self,
        updated_files: List[DynamicProxy],
        deleted_files: List[DynamicProxy],
        delete,
    ):
        modified_datasets = {}

        for file in updated_files:
            new_file = DatasetFile.from_path(
                client=self, path=file.entity.path, based_on=file.based_on, source=file.source
            )
            modified_datasets[file.dataset.name] = file.dataset
            file.dataset.add_or_update_files([new_file])

        if delete:
            for file in deleted_files:
                modified_datasets[file.dataset.name] = file.dataset
                file.dataset.unlink_file(file.entity.path)

        datasets_provenance = DatasetsProvenance()
        for dataset in modified_datasets.values():
            datasets_provenance.add_or_update(dataset, creator=Person.from_client(self))

    def update_dataset_git_files(self, files: List[DynamicProxy], ref, delete=False):
        """Update files and dataset metadata according to their remotes.

        :param files: List of files to be updated
        :param delete: Indicates whether to delete files or not

        :return: List of files that should be deleted
        """
        from renku import LocalClient

        visited_repos = {}
        updated_files: List[DynamicProxy] = []
        deleted_files: List[DynamicProxy] = []

        progress_text = "Checking files for updates"

        try:
            communication.start_progress(progress_text, len(files))
            for file in files:
                communication.update_progress(progress_text, 1)
                if not file.based_on:
                    continue

                based_on = file.based_on
                url = based_on.url
                if url in visited_repos:
                    repo, repo_path, remote_client = visited_repos[url]
                else:
                    repo, repo_path = self.prepare_git_repo(url, ref)
                    remote_client = LocalClient(repo_path)
                    visited_repos[url] = repo, repo_path, remote_client

                checksum = get_object_hash(repo, based_on.path)

                # NOTE: Use commit sha to compare if available for compatibility
                if based_on.commit_sha is not None:
                    remote_commit = find_previous_commit(repo, based_on.path)
                    found = bool(remote_commit)
                    changed = found and based_on.commit_sha != remote_commit.hexsha
                else:
                    found = bool(checksum)
                    changed = found and based_on.checksum != checksum

                src = Path(repo.working_dir) / based_on.path
                dst = self.renku_path.parent / file.entity.path

                if changed:
                    if src.exists():
                        # Fetch file if it is tracked by Git LFS
                        self._fetch_lfs_files(repo_path, {based_on.path})
                        if remote_client.is_external_file(src):
                            self.remove_file(dst)
                            self._create_external_file(src.resolve(), dst)
                        else:
                            shutil.copy(src, dst)
                        file.based_on = RemoteEntity(checksum=checksum, path=based_on.path, url=based_on.url)
                        updated_files.append(file)
                    else:
                        # File was removed or renamed
                        found = False

                if not found:
                    if delete:
                        self.remove_file(dst)
                    deleted_files.append(file)
        finally:
            communication.finalize_progress(progress_text)

        if not updated_files and (not delete or not deleted_files):
            # Nothing to commit or update
            return [], deleted_files

        # Commit changes in files

        file_paths = {str(self.path / f.entity.path) for f in updated_files + deleted_files}
        # Force-add to include possible ignored files that are in datasets
        add_to_git(self.repo.git, *file_paths, force=True)
        skip_hooks = not self.external_storage_requested
        self.repo.index.commit(
            "renku dataset: updated {} files and deleted {} files".format(len(updated_files), len(deleted_files)),
            skip_hooks=skip_hooks,
        )

        self._update_datasets_metadata(updated_files, deleted_files, delete)

        return updated_files, deleted_files

    def _create_external_file(self, src, dst):
        """Create a new external file."""
        try:
            pointer_file = self._create_pointer_file(target=src)
            relative = os.path.relpath(pointer_file, dst.parent)
            os.symlink(relative, dst)
        except OSError as e:
            raise errors.OperationError("Could not create symbolic link") from e

    def _create_pointer_file(self, target, checksum=None):
        """Create a new pointer file."""
        target = Path(target).resolve()

        if checksum is None:
            checksum = self._calculate_checksum(target)

        while True:
            filename = "{}-{}".format(uuid.uuid4(), checksum)
            path = self.renku_pointers_path / filename
            if not path.exists():
                break

        try:
            os.symlink(target, path)
        except FileNotFoundError:
            raise errors.ParameterError("Cannot find external file {}".format(target))

        return path

    def _calculate_checksum(self, filepath):
        try:
            return self.repo.git.hash_object(str(filepath))
        except GitCommandError:
            return None

    def update_external_files(self, records: List[DynamicProxy]):
        """Update files linked to external storage."""
        updated_files_paths = []
        updated_datasets = {}

        for file in records:
            if file.is_external:
                path = self.path / file.entity.path
                link = path.parent / os.readlink(path)
                pointer_file = self.path / link
                pointer_file = self._update_pointer_file(pointer_file)
                if pointer_file is not None:
                    relative = os.path.relpath(pointer_file, path.parent)
                    os.remove(path)
                    os.symlink(relative, path)
                    updated_files_paths.append(str(path))
                    updated_datasets[file.dataset.name] = file.dataset

        if not updated_files_paths:
            return

        add_to_git(self.repo.git, *updated_files_paths, force=True)
        self.repo.git.add(self.renku_pointers_path, force=True)
        self.repo.index.commit("renku dataset: updated {} external files".format(len(updated_files_paths)))

        datasets_provenance = DatasetsProvenance()

        for dataset in updated_datasets.values():
            for file in dataset.files:
                if str(self.path / file.entity.path) in updated_files_paths:
                    new_file = DatasetFile.from_path(client=self, path=file.entity.path, source=file.source)
                    dataset.add_or_update_files(new_file)

            datasets_provenance.add_or_update(dataset, creator=Person.from_client(self))

    def _update_pointer_file(self, pointer_file_path):
        """Update a pointer file."""
        try:
            target = pointer_file_path.resolve(strict=True)
        except FileNotFoundError:
            target = pointer_file_path.resolve()
            raise errors.ParameterError("External file not found: {}".format(target))

        checksum = self._calculate_checksum(target)
        current_checksum = pointer_file_path.name.split("-")[-1]

        if checksum == current_checksum:
            return

        os.remove(pointer_file_path)
        return self._create_pointer_file(target, checksum=checksum)

    @staticmethod
    def remove_file(filepath):
        """Remove a file/symlink and its pointer file (for external files)."""
        path = Path(filepath)
        try:
            link = path.parent / os.readlink(path)
        except FileNotFoundError:
            return
        except OSError:  # not a symlink but a normal file
            os.remove(path)
            return

        os.remove(path)

        try:
            os.remove(link)
        except FileNotFoundError:
            pass

    def is_external_file(self, path):
        """Checks if a path within repo is an external file."""
        path = self.path / path
        if not Path(path).is_symlink() or not self._is_path_within_repo(path):
            return False
        pointer = os.readlink(path)
        return f"{self.renku_home}/{self.POINTERS}" in pointer

    def has_external_files(self):
        """Return True if project has external files."""
        for dataset in self.datasets.values():
            for file_ in dataset.files:
                if file_.is_external:
                    return True

    def _is_path_within_repo(self, path):
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        path = Path(path)
        try:
            path.relative_to(self.path)
        except ValueError:
            return False
        else:
            return True

    def prepare_git_repo(self, url, ref=None, gitlab_token=None, renku_token=None, deployment_hostname=None, depth=1):
        """Clone and cache a Git repo."""
        if not url:
            raise errors.GitError("Invalid URL.")

        renku_branch = "renku-default-branch"

        depth = depth if not ref else None
        ref = ref or renku_branch
        u = GitURL.parse(url)
        path = u.path
        if u.host == "localhost":
            path = Path(path).resolve()
            git_url = str(path)
        elif "http" in u.scheme and gitlab_token:
            git_url = get_oauth_url(url, gitlab_token)
        elif "http" in u.scheme and renku_token:
            git_url = get_renku_repo_url(url, deployment_hostname=deployment_hostname, access_token=renku_token)
        else:
            git_url = url

        repo_name = os.path.splitext(os.path.basename(path))[0]
        path = os.path.dirname(path).lstrip("/")
        repo_path = self.renku_path / self.CACHE / u.host / path / repo_name

        if repo_path.exists():
            repo = Repo(str(repo_path))
            if have_same_remote(repo.remotes.origin.url, git_url):
                try:
                    repo.git.checkout(".")
                    repo.git.fetch(all=True)
                    repo.git.checkout(ref)
                    try:
                        repo.git.pull()
                    except GitError:
                        # When ref is not a branch, an error is thrown
                        pass
                except GitError:
                    # ignore the error and try re-cloning
                    pass
                else:
                    return repo, repo_path

            try:
                shutil.rmtree(str(repo_path))
            except PermissionError:
                raise errors.InvalidFileOperation(f"Cannot delete files in {repo_path}: Permission denied")

        repo, _ = clone(git_url, path=str(repo_path), install_githooks=False, depth=depth)

        # Because the name of the default branch is not always 'master', we
        # create an alias of the default branch when cloning the repo. It
        # is used to refer to the default branch later.
        renku_ref = "refs/heads/" + renku_branch
        try:
            repo.git.execute(["git", "symbolic-ref", renku_ref, repo.head.reference.path])
        except GitCommandError as e:
            raise errors.GitError(f"Cannot clone remote Git repo: {url}") from e

        try:
            repo.git.checkout(ref)
        except GitCommandError:
            raise errors.ParameterError(f"Cannot find reference '{ref}' in Git repository: {url}")

        return repo, repo_path

    def _download(self, url, filename, extract, chunk_size=16384):
        def extract_dataset(filepath):
            """Extract downloaded file."""
            try:
                tmp = tempfile.mkdtemp()
                patoolib.extract_archive(str(filepath), outdir=tmp, verbosity=-1)
            except patoolib.util.PatoolError:
                return filepath.parent, [filepath]
            else:
                filepath.unlink()
                return Path(tmp), [p for p in Path(tmp).rglob("*")]

        tmp_root = self.renku_path / self.CACHE
        tmp_root.mkdir(parents=True, exist_ok=True)
        tmp = tempfile.mkdtemp(dir=tmp_root)

        with requests.get(url, stream=True, allow_redirects=True) as request:
            request.raise_for_status()

            if not filename:
                filename = _filename_from_headers(request)

            if not filename:
                u = parse.urlparse(url)
                filename = Path(u.path).name
                if not filename:
                    raise errors.ParameterError(f"URL Cannot find a file to download from {url}")

            download_to = Path(tmp) / filename
            with open(str(download_to), "wb") as file_:
                total_size = int(request.headers.get("content-length", 0))

                communication.start_progress(name=filename, total=total_size)
                try:
                    for chunk in request.iter_content(chunk_size=chunk_size):
                        if chunk:  # ignore keep-alive chunks
                            file_.write(chunk)
                            communication.update_progress(name=filename, amount=len(chunk))
                finally:
                    communication.finalize_progress(name=filename)
        if extract:
            return extract_dataset(download_to)

        return download_to.parent, [download_to]


def _filename_from_headers(request):
    """Extract filename from content-disposition headers if available."""
    content_disposition = request.headers.get("content-disposition", None)

    if not content_disposition:
        return None

    entries = content_disposition.split(";")
    name_entry = next((e.strip() for e in entries if e.strip().lower().startswith("filename*=")), None)

    if name_entry:
        name = name_entry.split("=", 1)[1].strip()
        encoding, _, name = name.split("'")
        return unquote(name, encoding, errors="strict")

    name_entry = next((e.strip() for e in entries if e.strip().lower().startswith("filename=")), None)

    if not name_entry:
        return None

    filename = name_entry.split("=", 1)[1].strip()

    if filename.startswith('"'):
        filename = filename[1:-1]
    return filename


def _check_url(url):
    """Check if a url is local/remote and if it contains a git repository."""
    u = parse.urlparse(url)

    if u.scheme not in DatasetsApiMixin.SUPPORTED_SCHEMES:
        raise errors.UrlSchemeNotSupported('Scheme "{}" not supported'.format(u.scheme))

    is_remote = u.scheme not in ("", "file") or url.startswith("git@")
    is_git = False

    if is_remote:
        is_git = u.path.endswith(".git")
        if not is_git:
            # NOTE: Check if the url is a redirect.
            url = requests.head(url, allow_redirects=True).url
    else:
        try:
            Repo(u.path, search_parent_directories=True)
        except GitError:
            pass
        else:
            is_git = True

    return is_remote, is_git, url


DATASET_METADATA_PATHS = [
    Path(RENKU_HOME) / RepositoryApiMixin.DATABASE_PATH,
    Path(RENKU_HOME) / DatasetsApiMixin.DATASET_IMAGES,
    Path(RENKU_HOME) / DatasetsApiMixin.POINTERS,
    Path(RENKU_HOME) / LinkReference.REFS,
    ".gitattributes",
]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow management."""

import itertools
from collections import defaultdict
from typing import List, Set

import networkx

from renku.core.models.provenance.activity import Activity


def sort_activities(activities: List[Activity], remove_overridden_parents=True) -> List[Activity]:
    """Returns a sorted list of activities based on their dependencies and execution order."""
    by_usage = defaultdict(set)
    by_generation = defaultdict(set)

    overridden_activities = defaultdict(set)

    graph = networkx.DiGraph()

    def connect_nodes_based_on_dependencies():
        for activity in activities:
            # NOTE: Make sure that activity is in the graph in case it has no connection to others
            graph.add_node(activity)

            for usage in activity.usages:
                path = usage.entity.path
                by_usage[path].add(activity)
                parent_activities = by_generation[path]
                for parent in parent_activities:
                    create_edge(parent, activity, path)

            for generation in activity.generations:
                path = generation.entity.path
                by_generation[path].add(activity)
                child_activities = by_usage[path]
                for child in child_activities:
                    create_edge(activity, child, path)

    def create_edge(parent, child, path: str):
        if graph.has_edge(parent, child):
            return
        graph.add_edge(parent, child, path=path)

    def connect_nodes_by_execution_order():
        for path, values in by_generation.items():
            if len(values) <= 1:
                continue

            # NOTE: Order multiple activities that generate a common path
            create_order_among_activities(values, path)

    def create_order_among_activities(activities: Set[Activity], path):
        for a, b in itertools.combinations(activities, 2):
            if networkx.has_path(graph, a, b) or networkx.has_path(graph, b, a):
                continue

            # NOTE: More recent activity should be executed after the other one
            # NOTE: This won't introduce a cycle in the graph because there is no other path between the two nodes
            comparison = a.compare_to(b)
            if comparison < 0:
                graph.add_edge(a, b)
                overridden_activities[a].add(path)
            elif comparison > 0:
                graph.add_edge(b, a)
                overridden_activities[b].add(path)
            else:
                raise ValueError(f"Cannot create an order between activities {a.id} and {b.id}")

    def remove_overridden_activities():
        to_be_removed = set()
        to_be_processed = set(overridden_activities.keys())

        while len(to_be_processed) > 0:
            activity = to_be_processed.pop()
            overridden_paths = overridden_activities[activity]
            generated_path = {g.entity.path for g in activity.generations}
            if generated_path != overridden_paths:
                continue

            # NOTE: All generated paths are overridden; there is no point in executing the activity
            to_be_removed.add(activity)

            if not remove_overridden_parents:
                continue

            # NOTE: Check if its parents can be removed as well
            for parent in graph.predecessors(activity):
                if parent in to_be_removed:
                    continue
                data = graph.get_edge_data(parent, activity)
                if data and "path" in data:
                    overridden_activities[parent].add(data["path"])
                    to_be_processed.add(parent)

        for activity in to_be_removed:
            graph.remove_node(activity)

    connect_nodes_based_on_dependencies()

    if not networkx.algorithms.dag.is_directed_acyclic_graph(graph):
        raise ValueError("Cannot find execution order: Project has cyclic dependencies.")

    connect_nodes_by_execution_order()
    remove_overridden_activities()

    return list(networkx.topological_sort(graph))



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``rerun`` command."""

import os
import subprocess
from pathlib import Path

import git
import pytest
from click.testing import CliRunner

from renku.cli import cli
from tests.utils import format_result_exception, write_and_commit_file


def test_rerun(project, renku_cli):
    """Test rerun."""
    output = Path(project) / "output.txt"

    cmd = ["run", "python", "-S", "-c", "import random; print(random.random())"]

    assert 0 == renku_cli(*cmd, stdout=output).exit_code

    content = output.read_text().strip()

    def rerun():
        assert 0 == renku_cli("rerun", output).exit_code
        return output.read_text().strip()

    for _ in range(10):
        new_content = rerun()
        if content != new_content:
            break

    assert content != new_content, "Something is not random"


@pytest.mark.parametrize(
    "source, output",
    [
        ("input with space.txt", "output .txt"),
        ("coffee-orders-☕-by-location.csv", "😍works.txt"),
        ("test-愛", "成功"),
        ("그래프", "성공"),
        ("يحاول", "نجاح.txt"),
        ("график.txt", "успех.txt"),
        ("𒁃.c", "𒁏.txt"),
    ],
)
def test_rerun_with_special_paths(project, renku_cli, source, output):
    """Test rerun with unicode/whitespace filenames."""
    cwd = Path(project)
    source = cwd / source
    output = cwd / output

    assert 0 == renku_cli("run", "python", "-S", "-c", "import random; print(random.random())", stdout=source).exit_code
    assert 0 == renku_cli("run", "cat", source, stdout=output).exit_code

    content = output.read_text().strip()

    def rerun():
        assert 0 == renku_cli("rerun", output).exit_code
        return output.read_text().strip()

    for _ in range(10):
        new_content = rerun()
        if content != new_content:
            break

    assert content != output.read_text().strip(), "The output should have changed."


@pytest.mark.parametrize("source, content", [("input1", "input1 new-input2 old"), ("input2", "input1 old-input2 new")])
def test_rerun_with_from(project, renku_cli, source, content):
    """Test file recreation with specified inputs."""
    repo = git.Repo(project)
    cwd = Path(project)
    input1 = cwd / "input1"
    input2 = cwd / "input2"
    intermediate1 = cwd / "intermediate1"
    intermediate2 = cwd / "intermediate2"
    final1 = cwd / "final1"
    final2 = cwd / "final2"
    output = cwd / "output"

    write_and_commit_file(repo, input1, "input1 old-")
    write_and_commit_file(repo, input2, "input2 old")

    assert 0 == renku_cli("run", "cp", input1, intermediate1).exit_code
    assert 0 == renku_cli("run", "cp", input2, intermediate2).exit_code
    assert 0 == renku_cli("run", "cp", intermediate1, final1).exit_code
    assert 0 == renku_cli("run", "cp", intermediate2, final2).exit_code

    assert 0 == renku_cli("run", "cat", final1, final2, stdout=output).exit_code

    # Update both inputs
    write_and_commit_file(repo, input1, "input1 new-")
    write_and_commit_file(repo, input2, "input2 new")

    commit_sha_before = repo.head.object.hexsha

    assert 0 == renku_cli("rerun", "--from", source, output).exit_code

    assert content == output.read_text()

    commit_sha_after = repo.head.object.hexsha
    assert commit_sha_before != commit_sha_after


@pytest.mark.skip(reason="renku rerun not implemented with --edit-inputs yet, reenable later")
def test_rerun_with_edited_inputs(project, run, no_lfs_warning):
    """Test input modification."""
    runner = CliRunner(mix_stderr=False)

    cwd = Path(project)
    data = cwd / "examples"
    data.mkdir()
    first = data / "first.txt"
    second = data / "second.txt"
    third = data / "third.txt"

    run(args=["run", "echo", "hello"], stdout=first)
    run(args=["run", "cat", str(first)], stdout=second)
    run(args=["run", "echo", "1"], stdout=third)

    with first.open("r") as first_fp:
        with second.open("r") as second_fp:
            assert first_fp.read() == second_fp.read()

    # Change the initial input from "hello" to "hola".
    from click.testing import make_input_stream

    stdin = make_input_stream("hola\n", "utf-8")
    assert 0 == run(args=("rerun", "--edit-inputs", str(second)), stdin=stdin)

    with second.open("r") as second_fp:
        assert "hola\n" == second_fp.read()

    # Change the input from examples/first.txt to examples/third.txt.
    stdin = make_input_stream(str(third.name), "utf-8")
    old_dir = os.getcwd()
    try:
        # Make sure the input path is relative to the current directory.
        os.chdir(str(data))

        result = runner.invoke(
            cli, ["rerun", "--show-inputs", "--from", str(first), str(second)], catch_exceptions=False
        )
        assert 0 == result.exit_code, format_result_exception(result)
        assert result.output.startswith("https://")
        assert result.output[:-1].endswith(first.name)
        assert 0 == run(args=("rerun", "--edit-inputs", "--from", str(first), str(second)), stdin=stdin)
    finally:
        os.chdir(old_dir)

    with third.open("r") as third_fp:
        with second.open("r") as second_fp:
            assert third_fp.read() == second_fp.read()


def test_rerun_with_no_execution(project, runner):
    """Test rerun when no workflow is executed."""
    repo = git.Repo(project)
    input = os.path.join(project, "data", "input.txt")
    write_and_commit_file(repo, input, "content")

    result = runner.invoke(cli, ["rerun", input], catch_exceptions=False)

    assert 1 == result.exit_code
    assert "Path 'data/input.txt' is not generated by any workflows." in result.output


def test_output_directory(runner, project, run, no_lfs_size_limit):
    """Test detection of output directory."""
    cwd = Path(project)
    data = cwd / "source" / "data.txt"
    source = data.parent
    source.mkdir(parents=True)
    data.write_text("data")

    # Empty destination
    destination = cwd / "destination"
    source_wc = cwd / "destination_source.wc"
    # Non empty destination
    invalid_destination = cwd / "invalid_destination"
    invalid_destination.mkdir(parents=True)
    (invalid_destination / "non_empty").touch()

    repo = git.Repo(project)
    repo.git.add("--all")
    repo.index.commit("Created source directory", skip_hooks=True)

    cmd = ["run", "cp", "-LRf", str(source), str(destination)]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 0 == result.exit_code, format_result_exception(result)

    destination_source = destination / data.name
    assert destination_source.exists()

    # check that the output in subdir is added to LFS
    with (cwd / ".gitattributes").open() as f:
        gitattr = f.read()
    assert str(destination.relative_to(cwd)) + "/**" in gitattr
    assert destination_source.name in subprocess.check_output(["git", "lfs", "ls-files"]).decode()

    cmd = ["run", "wc"]
    assert 0 == run(args=cmd, stdin=destination_source, stdout=source_wc)

    # Make sure the output directory can be recreated
    assert 0 == run(args=("rerun", str(source_wc)))
    assert {data.name} == {path.name for path in destination.iterdir()}

    cmd = ["graph", "export"]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    destination_data = str(Path("destination") / "data.txt")
    assert destination_data in result.output, cmd

    cmd = ["run", "cp", "-r", str(source), str(invalid_destination)]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 1 == result.exit_code
    assert not (invalid_destination / data.name).exists()


def test_rerun_overridden_output(project, renku_cli, runner):
    """Test a path where final output is overridden won't be rerun."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r2", "cp", b, c]).exit_code
    assert 0 == renku_cli("run", "--name", "r3", "wc", a, stdout=c).exit_code

    result = runner.invoke(cli, ["rerun", "--dry-run", c])

    assert 0 == result.exit_code
    assert "r1" not in result.output
    assert "r2" not in result.output
    assert "r3" in result.output


def test_rerun_overridden_outputs_partially(project, renku_cli, runner):
    """Test a path where one of the final output is overridden won't be rerun."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")
    d = os.path.join(project, "d")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == renku_cli("run", "--name", "r2", "tee", c, d, stdin=b).exit_code
    assert 0 == renku_cli("run", "--name", "r3", "wc", a, stdout=c).exit_code

    result = runner.invoke(cli, ["rerun", "--dry-run", c])

    assert 0 == result.exit_code
    assert "r1" not in result.output
    assert "r2" not in result.output
    assert "r3" in result.output

    # Rerunning d will rerun r1 and r2
    result = runner.invoke(cli, ["rerun", "--dry-run", d])

    assert 0 == result.exit_code
    assert "r1" in result.output
    assert "r2" in result.output
    assert "r3" not in result.output


def test_rerun_multiple_paths_common_output(project, renku_cli, runner):
    """Test when multiple paths generate the same output only the most recent path will be rerun."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")
    d = os.path.join(project, "d")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r2", "cp", b, d]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r3", "cp", a, c]).exit_code
    assert 0 == renku_cli("run", "--name", "r4", "wc", c, stdout=d).exit_code

    result = runner.invoke(cli, ["rerun", "--dry-run", d])

    assert 0 == result.exit_code
    assert "r1" not in result.output
    assert "r2" not in result.output
    assert "r3" in result.output
    assert "r4" in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``project`` command."""

import json

from renku.cli import cli
from renku.core.metadata.gateway.project_gateway import ProjectGateway
from renku.core.models.provenance.agent import Person
from tests.utils import format_result_exception


def test_project_edit(runner, client, subdirectory, client_database_injection_manager):
    """Check project metadata editing."""
    (client.path / "README.md").write_text("Make repo dirty.")

    creator = "Forename Surname [Affiliation]"

    metadata = {
        "@id": "https://example.com/annotation1",
        "@type": "https://schema.org/specialType",
        "https://schema.org/specialProperty": "some_unique_value",
    }
    metadata_path = client.path / "metadata.json"
    metadata_path.write_text(json.dumps(metadata))

    commit_sha_before = client.repo.head.object.hexsha

    result = runner.invoke(
        cli, ["project", "edit", "-d", " new description ", "-c", creator, "--metadata", str(metadata_path)]
    )

    assert 0 == result.exit_code, format_result_exception(result)
    assert "Successfully updated: creator, description, custom_metadata." in result.output
    assert "Warning: No email or wrong format for: Forename Surname" in result.output

    with client_database_injection_manager(client):
        project_gateway = ProjectGateway()
        project = project_gateway.get_project()

    assert " new description " == project.description
    assert isinstance(project.creator, Person)
    assert "Forename Surname" == project.creator.name
    assert "Affiliation" == project.creator.affiliation
    assert metadata == project.annotations[0].body

    assert client.repo.is_dirty()
    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_before != commit_sha_after


def test_project_edit_no_change(runner, client):
    """Check project metadata editing does not commit when there is no change."""
    (client.path / "README.md").write_text("Make repo dirty.")

    commit_sha_before = client.repo.head.object.hexsha

    result = runner.invoke(cli, ["project", "edit"], catch_exceptions=False)

    assert 0 == result.exit_code, format_result_exception(result)
    assert "Nothing to update." in result.output

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after == commit_sha_before
    assert client.repo.is_dirty()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test `status` command."""

import os

from git import Repo

from renku.cli import cli
from tests.utils import format_result_exception, write_and_commit_file


def test_status(runner, project, subdirectory):
    """Test status check."""
    source = os.path.join(project, "source.txt")
    output = os.path.join(project, "data", "output.txt")

    repo = Repo(project)

    write_and_commit_file(repo, source, "content")

    assert 0 == runner.invoke(cli, ["run", "cp", source, output]).exit_code

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code, format_result_exception(result)

    write_and_commit_file(repo, source, "new content")

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code
    assert "Outdated outputs(1):" in result.output
    assert f"{os.path.relpath(output)}: {os.path.relpath(source)}" in result.output
    assert "Modified inputs(1):" in result.output
    assert "Outdated activities that have no outputs" not in result.output


def test_status_multiple_steps(runner, project):
    """Test status check with multiple steps."""
    source = os.path.join(os.getcwd(), "source.txt")
    intermediate = os.path.join(os.getcwd(), "intermediate.txt")
    output = os.path.join(os.getcwd(), "data", "output.txt")

    repo = Repo(project)

    write_and_commit_file(repo, source, "content")

    assert 0 == runner.invoke(cli, ["run", "cp", source, intermediate]).exit_code
    assert 0 == runner.invoke(cli, ["run", "cp", intermediate, output]).exit_code

    write_and_commit_file(repo, source, "new content")

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code
    assert "data/output.txt: source.txt" in result.output
    assert "intermediate.txt: source.txt" in result.output


def test_workflow_without_outputs(runner, project):
    """Test workflow without outputs."""
    source = os.path.join(os.getcwd(), "source.txt")

    repo = Repo(project)

    write_and_commit_file(repo, source, "content")

    result = runner.invoke(cli, ["run", "cat", "--no-output", source])
    assert 0 == result.exit_code, format_result_exception(result)

    write_and_commit_file(repo, source, "new content")

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code
    assert "Modified inputs(1):" in result.output
    assert "source.txt" in result.output
    assert "Outdated activities that have no outputs(1)" in result.output
    assert "/activities/" in result.output


def test_status_with_paths(runner, project):
    """Test status check with multiple steps."""
    source1 = os.path.join(os.getcwd(), "source1.txt")
    output1 = os.path.join(os.getcwd(), "data", "output1.txt")
    source2 = os.path.join(os.getcwd(), "source2.txt")
    output2 = os.path.join(os.getcwd(), "data", "output2.txt")

    repo = Repo(project)

    write_and_commit_file(repo, source1, "content")
    write_and_commit_file(repo, source2, "content")

    assert 0 == runner.invoke(cli, ["run", "cp", source1, output1]).exit_code
    assert 0 == runner.invoke(cli, ["run", "cp", source2, output2]).exit_code

    write_and_commit_file(repo, source1, "new content")
    write_and_commit_file(repo, source2, "new content")

    result = runner.invoke(cli, ["status", source1])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output1.txt: source1.txt" in result.output
    assert "Modified inputs(1):" in result.output
    assert "source2.txt" not in result.output

    result = runner.invoke(cli, ["status", output1])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output1.txt: source1.txt" in result.output
    assert "Modified inputs(1):" in result.output
    assert "source2.txt" not in result.output

    result = runner.invoke(cli, ["status", "source2.txt"])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output2.txt: source2.txt" in result.output
    assert "Modified inputs(1):" in result.output
    assert "source1.txt" not in result.output

    result = runner.invoke(cli, ["status", "data/output2.txt"])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output2.txt: source2.txt" in result.output
    assert "Modified inputs(1):" in result.output
    assert "source1.txt" not in result.output

    result = runner.invoke(cli, ["status", source1, output2])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output1.txt: source1.txt" in result.output
    assert "data/output2.txt: source2.txt" in result.output
    assert "Modified inputs(2):" in result.output


def test_status_with_path_all_generation(runner, project):
    """Test that all generations are reported if only one of them is specified."""
    source = os.path.join(project, "source.txt")
    output1 = os.path.join(project, "data", "output1.txt")
    output2 = os.path.join(project, "data", "output2.txt")

    repo = Repo(project)

    write_and_commit_file(repo, source, "content")

    assert 0 == runner.invoke(cli, ["run", "--input", source, "touch", output1, output2]).exit_code

    write_and_commit_file(repo, source, "new content")

    result = runner.invoke(cli, ["status", output1])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "data/output1.txt: source.txt" in result.output
    assert "data/output2.txt: source.txt" in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku CLI fixtures for execution management."""

from collections import namedtuple
from typing import List, Tuple, Union

import pytest

Result = namedtuple("Result", "exit_code, activities")


@pytest.fixture
def renku_cli(client, run, client_database_injection_manager):
    """Return a callable Renku CLI.

    It returns the exit code and the resulting activity or list of activities.
    """
    from renku.core.management.command_builder.command import inject
    from renku.core.management.interface.activity_gateway import IActivityGateway
    from renku.core.models.provenance.activity import Activity

    def renku_cli_(*args, **kwargs) -> Tuple[int, Union[None, Activity, List[Activity]]]:
        @inject.autoparams()
        def _get_activities(activity_gateway: IActivityGateway):
            return {a.id: a for a in activity_gateway.get_latest_activity_per_plan().values()}

        with client_database_injection_manager(client):
            activities_before = _get_activities()

        args = [str(a) for a in args]

        exit_code = run(args, **kwargs)

        with client_database_injection_manager(client):
            activities_after = _get_activities()

        new_activities = [a for id, a in activities_after.items() if id not in activities_before]

        if len(new_activities) == 0:
            new_activities = None
        elif len(new_activities) == 1:
            new_activities = new_activities[0]

        return Result(exit_code, new_activities)

    return renku_cli_



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test plugins for the ``workflow`` command."""

from renku.cli import cli
from renku.core.plugins import pluginmanager as pluginmanager


def test_renku_workflow_exporter_hook(monkeypatch, dummy_workflow_exporter_hook, runner, project):
    """Tests that the renku workflow export plugin hook on ``Plan`` is called."""
    pm = pluginmanager.get_plugin_manager()
    pm.register(dummy_workflow_exporter_hook)

    with monkeypatch.context() as m:
        m.setattr(pluginmanager, "get_plugin_manager", lambda: pm)

        cmd = ["echo", "test"]
        run_name = "run_1"
        result = runner.invoke(cli, ["run", "--no-output", "--name", run_name] + cmd)
        assert 0 == result.exit_code

        result = runner.invoke(cli, ["workflow", "export", "--format", "dummy", run_name])
        assert "dummy" in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku database dispatcher interface."""

from abc import ABC

from renku.core.metadata.database import Database


class IDatabaseDispatcher(ABC):
    """Interface for the DatabaseDispatcher.

    Handles getting current database (Database) and entering/exiting the stack for the database.
    """

    @property
    def current_database(self) -> Database:
        """Get the currently active database."""
        raise NotImplementedError

    def push_database_to_stack(self, path: str, commit: bool = False) -> None:
        """Create and push a new database to the stack."""
        raise NotImplementedError

    def pop_database(self) -> None:
        """Remove the current database from the stack."""
        raise NotImplementedError

    def finalize_dispatcher(self) -> None:
        """Close all database contexts."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Immutable classes."""

import pytest

from renku.core.metadata.immutable import Immutable, Slots


class A(Slots):
    """Test class."""

    __slots__ = ("a_member",)


class B(A):
    """Test class."""

    __slots__ = "b_member"

    def __init__(self, *, b_member=None, **kwargs):
        super().__init__(b_member=b_member, **kwargs)


class C(Immutable):
    """Test class."""

    __slots__ = ("c_member",)


def test_instantiate():
    """Test instantiating Slots subclasses."""
    b = B(a_member=42, b_member=43)

    assert {"a_member": 42, "b_member": 43} == b.__getstate__()


def test_instantiate_incomplete():
    """Test instantiating Slots subclasses without setting all members."""
    b = B(a_member=42)

    assert {"a_member": 42, "b_member": None} == b.__getstate__()


def test_instantiate_invalid_member():
    """Test instantiating Slots subclasses and passing a non-member."""
    with pytest.raises(AttributeError) as e:
        B(c_member=42)

    assert "object has no attribute 'c_member'" in str(e)


def test_get_all_slots():
    """Test get all slots from an Immutable subclasses."""
    b = B(a_member=42, b_member=43)
    _ = b.__getstate__()

    assert {"b_member", "a_member", "__weakref__"} == set(B.__all_slots__)


def test_immutable_object_id():
    """Test Immutable subclasses have an `id` field."""
    c = C(id=42, c_member=43)

    assert {"c_member": 43, "id": 42} == c.__getstate__()


def test_cannot_mutate():
    """Test cannot mutate an Immutable subclasses."""
    c = C(c_member=42)

    with pytest.raises(TypeError) as e:
        c.c_member = None

    assert "Cannot modify an immutable class" in str(e)
    assert 42 == c.c_member


def test_immutable_objects_cache():
    """Test Immutable objects are cached once created."""
    data = {"id": 42, "c_member": 43}

    o1 = C.make_instance(C(**data))
    o2 = C.make_instance(C(**data))

    assert o1 is o2


def test_immutable_objects_cache_without_id():
    """Test Immutable objects cannot be cached if id is not set."""
    data = {"c_member": 43}

    o1 = C.make_instance(C(**data))
    o2 = C.make_instance(C(**data))

    assert o1 is not o2



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git utility functions."""

import math
import pathlib
import subprocess
import tempfile
import urllib
from pathlib import Path
from subprocess import SubprocessError, run
from typing import Optional, Union

from git import Commit, Git, GitCommandError, Repo

from renku.core import errors
from renku.core.models.git import GitURL

ARGUMENT_BATCH_SIZE = 100
CLI_GITLAB_ENDPOINT = "repos"


def run_command(command, *paths, separator=None, **kwargs):
    """Execute command by splitting `paths` to make sure that argument list will be within os limits.

    :param command: A list or tuple containing command and its arguments.
    :param paths: List of paths/long argument. This will be appended to `command` for each invocation.
    :param separator: Separator for `paths` if they need to be passed as string.
    :param kwargs: Extra arguments passed to `subprocess.run`.
    :returns: Result of last invocation.
    """
    result = None

    for batch in split_paths(*paths):
        if separator:
            batch = [separator.join(batch)]

        try:
            if not isinstance(batch, list):
                batch = list(batch)
            result = run(command + batch, **kwargs)

            if result.returncode != 0:
                break
        except KeyboardInterrupt:
            raise
        except SubprocessError as e:
            raise errors.GitError(f"Cannot run command {command} : {e}")

    return result


def add_to_git(git, *paths, **kwargs):
    """Split `paths` and add them to git to make sure that argument list will be within os limits."""
    for batch in split_paths(*paths):
        git.add(*batch, **kwargs)


def split_paths(*paths):
    """Return a generator with split list of paths."""
    batch_count = math.ceil(len(paths) / ARGUMENT_BATCH_SIZE)
    batch_count = max(batch_count, 1)

    for index in range(batch_count):
        yield paths[index * ARGUMENT_BATCH_SIZE : (index + 1) * ARGUMENT_BATCH_SIZE]


def get_oauth_url(url, gitlab_token):
    """Format URL with a username and password."""
    parsed_url = urllib.parse.urlparse(url)

    if not parsed_url.netloc:
        raise ValueError(f"Invalid http git url: {url}")

    netloc = f"oauth2:{gitlab_token}@{parsed_url.netloc}"
    return parsed_url._replace(netloc=netloc).geturl()


def have_same_remote(url1, url2):
    """Checks if two git urls point to the same remote repo ignoring protocol and credentials."""
    u1 = GitURL.parse(url1)
    u2 = GitURL.parse(url2)

    return u1.hostname == u2.hostname and u1.pathname == u2.pathname


def get_renku_repo_url(remote_url, deployment_hostname=None, access_token=None):
    """Return a repo url that can be authenticated by renku."""
    parsed_remote = GitURL.parse(remote_url)
    path = parsed_remote.pathname.strip("/")
    if path.startswith("gitlab/"):
        path = path.replace("gitlab/", "")
    path = pathlib.posixpath.join(CLI_GITLAB_ENDPOINT, path)

    credentials = f"renku:{access_token}@" if access_token else ""
    hostname = deployment_hostname or parsed_remote.hostname

    return urllib.parse.urljoin(f"https://{credentials}{hostname}", path)


def get_object_hash(repo: Repo, path: Union[Path, str], revision: str = None) -> Optional[str]:
    """Return git hash of an object in a Repo or its submodule.

    NOTE: path must be relative to the repo's root regardless if this function is called from a subdirectory or not.
    """

    def get_object_hash_from_submodules() -> Optional[str]:
        for submodule in repo.submodules:
            try:
                path_in_submodule = Path(path).relative_to(submodule.path)
            except ValueError:
                continue
            else:
                try:
                    return Git(submodule.abspath).rev_parse(f"HEAD:{str(path_in_submodule)}")
                except GitCommandError:
                    pass

    revision = revision or "HEAD"

    try:
        return repo.git.rev_parse(f"{revision}:{str(path)}")
    except GitCommandError:
        # NOTE: The file can be in a submodule or it was not there when the command ran but was there when workflows
        # were migrated (this can happen only for Usage); the project might be broken too.
        return get_object_hash_from_submodules()


def get_content(repo: Repo, path: Union[Path, str], checksum: str) -> str:
    """Get content of an object from git using its checksum and return a path with that content."""

    def get_content_from_submodules(file):
        for submodule in repo.submodules:
            try:
                path_in_submodule = Path(path).relative_to(submodule.path)
            except ValueError:
                continue
            else:
                try:
                    subprocess.run(
                        ["git", "cat-file", "--filters", "--path", str(path_in_submodule), checksum],
                        check=True,
                        stdout=file,
                        cwd=repo.working_dir,
                    )
                except subprocess.CalledProcessError:
                    pass
                else:
                    return True

    with tempfile.NamedTemporaryFile(mode="w+b", delete=False) as temp_file:
        try:
            subprocess.run(
                ["git", "cat-file", "--filters", "--path", str(path), checksum],
                check=True,
                stdout=temp_file,
                cwd=repo.working_dir,
            )
        except subprocess.CalledProcessError:
            if not get_content_from_submodules(temp_file):
                raise errors.ExportError(f"File not found in repo: {checksum}:{path}")

    return temp_file.name


def find_previous_commit(
    repo: Repo, path: Union[Path, str], revision: str = None, return_first=False, full_history=False
) -> Optional[Commit]:
    """Return a previous commit for a given path starting from ``revision``.

    :param path: relative path to the file
    :param revision: revision to start from, defaults to ``HEAD``
    :param return_first: show the first commit in the history
    :param full_history: search full history
    """

    def get_previous_commit_from_submodules() -> Optional[Commit]:
        for submodule in repo.submodules:
            try:
                path_in_submodule = Path(path).relative_to(submodule.path)
            except ValueError:
                continue
            else:
                try:
                    return find_previous_commit(
                        repo=Repo(submodule.abspath),
                        path=path_in_submodule,
                        revision=revision,
                        return_first=return_first,
                        full_history=full_history,
                    )
                except KeyError:
                    pass

    path = str(path)
    revision = revision or "HEAD"
    kwargs = {}

    if full_history:
        kwargs["full_history"] = True
    if not return_first:
        kwargs["max_count"] = 1

    commits = list(repo.iter_commits(revision, paths=path, **kwargs))

    if not commits:
        commit = get_previous_commit_from_submodules()
        return commit

    return commits[-1 if return_first else 0]


def get_path(url: str):
    """Return path part of a url."""
    return urllib.parse.urlparse(url).path



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test plan database gateways."""

from datetime import datetime

from renku.core.metadata.gateway.plan_gateway import PlanGateway
from renku.core.models.workflow.composite_plan import CompositePlan
from renku.core.models.workflow.plan import Plan


def test_plan_gateway_add_get(dummy_database_injection_manager):
    """Test getting a plan by id."""

    plan = Plan(id=Plan.generate_id(), name="plan")
    composite_plan = CompositePlan(id=CompositePlan.generate_id(), name="composite-plan")

    with dummy_database_injection_manager(None):
        plan_gateway = PlanGateway()

        plan_gateway.add(plan)
        plan_gateway.add(composite_plan)

        assert plan == plan_gateway.get_by_id(plan.id)
        assert composite_plan == plan_gateway.get_by_id(composite_plan.id)

        assert plan == plan_gateway.get_by_name("plan")
        assert composite_plan == plan_gateway.get_by_name("composite-plan")

        assert not plan_gateway.get_by_name(plan.id)
        assert not plan_gateway.get_by_name(composite_plan.id)

        assert not plan_gateway.get_by_id("plan")
        assert not plan_gateway.get_by_id("composite-plan")


def test_plan_gateway_newest_plans(dummy_database_injection_manager, database):
    """Test getting newest plans."""
    plan = Plan(id=Plan.generate_id(), name="plan")
    plan2 = Plan(id=Plan.generate_id(), name="plan")
    invalidated_plan = Plan(id=Plan.generate_id(), name="invalidated_plan", invalidated_at=datetime.utcnow())
    invalidated_plan2 = Plan(id=Plan.generate_id(), name="invalidated_plan", invalidated_at=datetime.utcnow())

    with dummy_database_injection_manager(None):
        plan_gateway = PlanGateway()

        plan_gateway.add(plan)
        plan_gateway.add(plan2)
        plan_gateway.add(invalidated_plan)
        plan_gateway.add(invalidated_plan2)

        newest_plans_by_names = {p.id for p in plan_gateway.get_newest_plans_by_names().values()}

        assert {plan2.id} == newest_plans_by_names

        newest_plans_by_names_with_invalidated = {
            p.id for p in plan_gateway.get_newest_plans_by_names(with_invalidated=True).values()
        }

        assert {plan2.id, invalidated_plan2.id} == newest_plans_by_names_with_invalidated



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow commands tests."""


from contextlib import nullcontext

import pytest

from renku.core import errors
from renku.core.management.workflow.concrete_execution_graph import ExecutionGraph
from renku.core.management.workflow.value_resolution import CompositePlanValueResolver
from renku.core.models.workflow.composite_plan import CompositePlan


def _get_nested_actual_values(run):
    """Get a dict of parameter -> actual values mappings from a run."""
    result = dict()
    if isinstance(run, CompositePlan):
        for step in run.plans:
            result[step.name] = _get_nested_actual_values(step)
    else:
        for param in run.inputs + run.outputs + run.parameters:
            result[param.name] = param.actual_value

    return result


def test_composite_plan_resolve_relative_mapping(composite_plan):
    """Test if resolving a relative command parameter mapping works."""

    grouped, run1, run2 = composite_plan

    assert grouped.resolve_mapping_path("@step1.@input1")[0] == run1.inputs[0]
    assert grouped.resolve_mapping_path("@step1.@input2")[0] == run1.inputs[1]
    assert grouped.resolve_mapping_path("@step1.@output1")[0] == run1.outputs[0]
    assert grouped.resolve_mapping_path("@step1.@output2")[0] == run1.outputs[1]
    assert grouped.resolve_mapping_path("@step1.@param1")[0] == run1.parameters[0]
    assert grouped.resolve_mapping_path("@step1.@param2")[0] == run1.parameters[1]

    assert grouped.resolve_mapping_path("@step2.@input1")[0] == run2.inputs[0]
    assert grouped.resolve_mapping_path("@step2.@input2")[0] == run2.inputs[1]
    assert grouped.resolve_mapping_path("@step2.@output1")[0] == run2.outputs[0]
    assert grouped.resolve_mapping_path("@step2.@output2")[0] == run2.outputs[1]
    assert grouped.resolve_mapping_path("@step2.@param1")[0] == run2.parameters[0]
    assert grouped.resolve_mapping_path("@step2.@param2")[0] == run2.parameters[1]


def test_composite_plan_resolve_absolute_mapping(composite_plan):
    """Test if resolving an absolute command parameter mapping works."""

    grouped, run1, run2 = composite_plan

    assert grouped.resolve_mapping_path("run1.run1_input1")[0] == run1.inputs[0]
    assert grouped.resolve_mapping_path("run1.run1_input2")[0] == run1.inputs[1]
    assert grouped.resolve_mapping_path("run1.run1_output1")[0] == run1.outputs[0]
    assert grouped.resolve_mapping_path("run1.run1_output2")[0] == run1.outputs[1]
    assert grouped.resolve_mapping_path("run1.run1_param1")[0] == run1.parameters[0]
    assert grouped.resolve_mapping_path("run1.run1_param2")[0] == run1.parameters[1]

    assert grouped.resolve_mapping_path("run2.run2_input1")[0] == run2.inputs[0]
    assert grouped.resolve_mapping_path("run2.run2_input2")[0] == run2.inputs[1]
    assert grouped.resolve_mapping_path("run2.run2_output1")[0] == run2.outputs[0]
    assert grouped.resolve_mapping_path("run2.run2_output2")[0] == run2.outputs[1]
    assert grouped.resolve_mapping_path("run2.run2_param1")[0] == run2.parameters[0]
    assert grouped.resolve_mapping_path("run2.run2_param2")[0] == run2.parameters[1]


@pytest.mark.parametrize(
    "mappings, names",
    [
        (["prop1=@step1.@input1"], {"prop1"}),
        (
            ["prop_abcdefg123A=run2.run2_output1", "prop47=run2.@output2", "something=@step1.run1_param1"],
            {"prop_abcdefg123A", "prop47", "something"},
        ),
        (["prop1=@step1.@input1", "prop2=prop1"], {"prop1", "prop2"}),
    ],
)
def test_composite_plan_create_mapping(composite_plan, mappings, names):
    """Test parsing of mapping strings."""
    grouped, _, _ = composite_plan

    grouped.set_mappings_from_strings(mappings)

    created_mappings = {m.name for m in grouped.mappings}

    assert created_mappings == names


def test_composite_plan_create_mapping_nested(composite_plan):
    """Test parsing of mapping strings."""
    grouped, run1, _ = composite_plan

    grouped.set_mappings_from_strings(["prop1=@step1.@input1", "prop2=@step2.@output2"])

    grouped2 = CompositePlan(id=CompositePlan.generate_id(), plans=[run1, grouped], name="grouped2")

    with pytest.raises(errors.MappingExistsError):
        grouped2.set_mappings_from_strings(["prop2=grouped1.@mapping1", "prop3=grouped1.prop1"])

    grouped2.set_mappings_from_strings(["prop3=grouped1.prop2", "direct=run1.@input1", "nested=grouped1.run1.@input2"])

    assert len(grouped2.mappings) == 4

    assert next(filter(lambda m: m.name == "prop2", grouped2.mappings)).mapped_parameters[0] == grouped.mappings[0]
    assert next(filter(lambda m: m.name == "prop3", grouped2.mappings)).mapped_parameters[0] == grouped.mappings[1]
    assert next(filter(lambda m: m.name == "direct", grouped2.mappings)).mapped_parameters[0] == run1.inputs[0]
    assert next(filter(lambda m: m.name == "nested", grouped2.mappings)).mapped_parameters[0] == run1.inputs[1]


def test_composite_plan_set_defaults(composite_plan):
    """Test setting default values on grouping runs."""

    grouped, run1, run2 = composite_plan

    grouped.set_mappings_from_strings(["prop1=@step1.@input1", "prop2=run2.@output2"])

    grouped.set_mapping_defaults(["prop1=1", "prop2=abcdefg", "@step1.@param1=xyz", "run2.run2_input1=7.89"])

    assert next(filter(lambda m: m.name == "prop1", grouped.mappings)).default_value == "1"
    assert next(filter(lambda m: m.name == "prop2", grouped.mappings)).default_value == "abcdefg"
    assert run1.parameters[0].default_value == "xyz"
    assert run2.inputs[0].default_value == "7.89"

    assert run1.inputs[0].default_value != "1"
    assert run2.outputs[1].default_value != "abcdefg"


def test_composite_plan_set_description(composite_plan):
    """Test setting description on grouping runs."""

    grouped, run1, run2 = composite_plan

    grouped.set_mappings_from_strings(["prop1=@step1.@input1", "prop2=run2.@output2"])

    grouped.set_mapping_descriptions(['prop1="first property"', 'prop2="another property?"'])

    assert next(filter(lambda m: m.name == "prop1", grouped.mappings)).description == "first property"
    assert next(filter(lambda m: m.name == "prop2", grouped.mappings)).description == "another property?"

    with pytest.raises(errors.ParameterNotFoundError):
        grouped.set_mapping_descriptions(["@step1.@param1=xyz", "run2.run2_input1=7.89"])


def test_composite_plan_map_all_inputs(composite_plan):
    """Test automatic mapping of all child inputs."""

    grouped, run1, run2 = composite_plan

    grouped.map_all_inputs()
    assert len(grouped.mappings) == 4
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.inputs[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.inputs[1])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.inputs[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.inputs[1])


def test_composite_plan_map_all_outputs(composite_plan):
    """Test automatic mapping of all child outputs."""

    grouped, run1, run2 = composite_plan

    grouped.map_all_outputs()
    assert len(grouped.mappings) == 4
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.outputs[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.outputs[1])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.outputs[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.outputs[1])


def test_composite_plan_map_all_parameters(composite_plan):
    """Test automatic mapping of all child parameters."""

    grouped, run1, run2 = composite_plan

    grouped.map_all_parameters()
    assert len(grouped.mappings) == 4
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.parameters[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run1.parameters[1])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.parameters[0])
    assert any(m for m in grouped.mappings if m.mapped_parameters[0] == run2.parameters[1])


@pytest.mark.parametrize(
    "mappings, defaults, values, expected",
    [
        # NOTE: Test without anything
        (
            [],
            [],
            {},
            {
                "run1": {
                    "run1_input1": 1,
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": 6,
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": 2,
                    "run2_output1": 3,
                    "run2_output2": 4,
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with values applied
        (
            [],
            [],
            {
                "steps": {
                    "run1": {
                        "run1_param2": "f",
                    },
                    "run2": {
                        "run2_input2": "h",
                    },
                }
            },
            {
                "run1": {
                    "run1_input1": 1,
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": "f",
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": "h",
                    "run2_output1": 3,
                    "run2_output2": 4,
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with mappings
        (
            ["m1=@step1.@input1", "m2=@step2.@output2"],
            [],
            {},
            {
                "run1": {
                    "run1_input1": 1,
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": 6,
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": 2,
                    "run2_output1": 3,
                    "run2_output2": 4,
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with mappings and defaults
        (
            ["m1=@step1.@input1", "m2=@step2.@output2"],
            ["m1=x", "m2=y"],
            {},
            {
                "run1": {
                    "run1_input1": "x",
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": 6,
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": 2,
                    "run2_output1": 3,
                    "run2_output2": "y",
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with mappings, defaults and values
        (
            ["m1=@step1.@input1", "m2=@step2.@output2"],
            ["m1=x", "m2=y"],
            {
                "steps": {
                    "run1": {
                        "run1_param2": "f",
                    },
                    "run2": {"run2_input2": "h", "run2_output2": "a"},  # NOTE: Override value provided by mapping
                }
            },
            {
                "run1": {
                    "run1_input1": "x",
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": "f",
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": "h",
                    "run2_output1": 3,
                    "run2_output2": "a",
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with mappings, defaults and values for mappings and params
        (
            ["m1=@step1.@input1", "m2=@step2.@output2"],
            ["m1=x", "m2=y"],
            {
                "parameters": {"m2": "z"},
                "steps": {
                    "run1": {
                        "run1_param2": "f",
                    },
                    "run2": {"run2_input2": "h"},  # NOTE: Override value provided by mapping
                },
            },
            {
                "run1": {
                    "run1_input1": "x",
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": "f",
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": "h",
                    "run2_output1": 3,
                    "run2_output2": "z",
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
        # NOTE: Test with mappings, defaults and values for mappings and params
        (
            ["m1=@step1.@input1", "m2=@step2.@output2"],
            ["m1=x", "m2=y"],
            {
                "parameters": {"m1": "42"},
                "steps": {
                    "run1": {
                        "run1_param2": "f",
                    },
                    "run2": {"run2_input2": "h", "run2_output2": "a"},  # NOTE: Override value provided by mapping
                },
            },
            {
                "run1": {
                    "run1_input1": "42",
                    "run1_input2": 2,
                    "run1_output1": 3,
                    "run1_output2": 4,
                    "run1_param1": 5,
                    "run1_param2": "f",
                },
                "run2": {
                    "run2_input1": 1,
                    "run2_input2": "h",
                    "run2_output1": 3,
                    "run2_output2": "a",
                    "run2_param1": 5,
                    "run2_param2": 6,
                },
            },
        ),
    ],
)
def test_composite_plan_actual_values(composite_plan, mappings, defaults, values, expected):
    """Test resolving actual values on a grouped run."""

    grouped, _, _ = composite_plan

    grouped.set_mappings_from_strings(mappings)
    grouped.set_mapping_defaults(defaults)
    rv = CompositePlanValueResolver(grouped, values)

    actual = _get_nested_actual_values(rv.apply())
    assert len(rv.missing_parameters) == 0

    assert actual == expected


@pytest.mark.parametrize(
    "links,raises,cycles",
    [
        ([], False, False),
        (["@mapping1=@mapping2"], True, False),
        (["@mapping1=@step1.@input1"], True, False),
        (["@step1.@input1=@step2.@output2"], True, False),
        (["@step2.@output2=@step1.@input1"], False, False),
        (["@step2.@output2=@step1.@input1", "@step2.@output1=@step1.@input1"], True, False),
        (["@step2.@output2=@step1.@input1", "@step1.@output1=@step2.@input1"], False, True),
    ],
)
def test_composite_plan_links(composite_plan, links, raises, cycles):
    """Test adding links to grouped runs."""

    grouped, _, _ = composite_plan
    grouped.set_mappings_from_strings(["prop1=@step1.@input1", "prop2=run2.@output2"])

    if raises:
        with pytest.raises(errors.ParameterLinkError):
            grouped.set_links_from_strings(links)
    else:
        grouped.set_links_from_strings(links)
        found_cycles = ExecutionGraph(grouped).cycles
        assert bool(found_cycles) == cycles, found_cycles


@pytest.mark.parametrize(
    "mappings,defaults,links,raises,cycles",
    [
        ([], [], False, False, False),
        ([], ["@step1.@output1=myfile.txt", "@step2.@input1=myfile.txt"], True, False, False),
        (
            [],
            ["@step1.@output1=myfile.txt", "@step2.@input1=myfile.txt", "@step1.@input1=myfile.txt"],
            True,
            True,
            True,
        ),
        (
            [],
            [
                "@step1.@output1=myfile.txt",
                "@step2.@input1=myfile.txt",
                "@step1.@input1=otherfile.txt",
                "@step2.@output2=otherfile.txt",
            ],
            True,
            False,
            True,
        ),
        (
            ["input=@step1.@input1", "output=@step2.@output2"],
            ["input=myfile.txt", "output=myfile.txt"],
            True,
            False,
            False,
        ),
        (
            ["input=@step1.@input1,@step2.@input2", "output=@step2.@output2"],
            ["input=myfile.txt", "output=myfile.txt"],
            True,
            True,
            True,
        ),
        (["map=@step1.@input1,@step2.@input2,@step1.@output1,@step2.@output2"], ["map=myfile.txt"], True, True, True),
    ],
)
def test_composite_plan_auto_links(composite_plan, mappings, defaults, links, raises, cycles):
    """Test automatically detecting links between steps."""
    grouped, _, _ = composite_plan

    if raises:
        maybe_raises = pytest.raises(errors.ParameterLinkError)
    else:
        maybe_raises = nullcontext()

    grouped.set_mappings_from_strings(mappings)
    grouped.set_mapping_defaults(defaults)

    rv = CompositePlanValueResolver(grouped, None)
    grouped = rv.apply()
    assert len(rv.missing_parameters) == 0

    graph = ExecutionGraph(grouped, virtual_links=True)

    assert bool(graph.virtual_links) == links
    assert bool(graph.cycles) == cycles

    with maybe_raises:
        for virtual_link in graph.virtual_links:
            grouped.add_link(virtual_link[0], [virtual_link[1]])



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service fixtures for project management."""
import os
import tempfile
import uuid
from pathlib import Path
from urllib.parse import urlparse

import git
import pytest

from renku.core.utils.scm import normalize_to_ascii


@pytest.fixture
def project_metadata(project):
    """Create project with metadata."""
    name = Path(project).name
    metadata = {
        "project_id": uuid.uuid4().hex,
        "name": name,
        "slug": normalize_to_ascii(name),
        "fullname": "full project name",
        "email": "my@email.com",
        "owner": "me",
        "token": "awesome token",
        "git_url": "git@gitlab.com",
        "initialized": True,
    }

    yield project, metadata


@pytest.fixture(scope="module")
def it_git_access_token():
    """Returns a git access token for a testing run."""
    from tests.fixtures.config import IT_GIT_ACCESS_TOKEN

    return IT_GIT_ACCESS_TOKEN


@pytest.fixture(scope="module")
def it_remote_repo_url():
    """Returns a remote path to integration test repository."""
    from tests.fixtures.config import IT_REMOTE_REPO_URL

    return IT_REMOTE_REPO_URL


@pytest.fixture(scope="function")
def it_remote_repo_url_temp_branch(it_remote_repo_url):
    """Returns a remote path to integration test repository."""

    with tempfile.TemporaryDirectory() as tempdir:
        # NOTE: create temporary branch and push it
        git_url = urlparse(it_remote_repo_url)

        url = "oauth2:{0}@{1}".format(os.getenv("IT_OAUTH_GIT_TOKEN"), git_url.netloc)
        git_url = git_url._replace(netloc=url).geturl()
        repo = git.Repo.clone_from(git_url, tempdir)
        origin = repo.remote("origin")
        branch_name = uuid.uuid4().hex
        branch = repo.create_head(branch_name)
        repo.git.push("--set-upstream", origin, branch)
        try:
            yield it_remote_repo_url, branch_name
        finally:
            # NOTE: delete remote branch
            origin.push(f":{branch}")
            repo.delete_head(branch)


@pytest.fixture(scope="module")
def it_non_renku_repo_url():
    """Returns a remote path to integration test repository."""
    from tests.fixtures.config import IT_REMOTE_NON_RENKU_REPO_URL

    return IT_REMOTE_NON_RENKU_REPO_URL


@pytest.fixture(scope="module")
def it_no_commit_repo_url():
    """Returns a remote path to integration test repository."""
    from tests.fixtures.config import IT_REMOTE_NO_COMMITS_REPO_URL

    return IT_REMOTE_NO_COMMITS_REPO_URL


@pytest.fixture(scope="module")
def it_protected_repo_url():
    """Returns a protected repository url."""
    from tests.fixtures.config import IT_PROTECTED_REMOTE_REPO_URL

    return IT_PROTECTED_REMOTE_REPO_URL
